// -*- mode: c++; mode: fold -*-
// Description                              /*{{{*/
// $Id: dudf.cc,v 1.2 2002/07/25 18:07:18 prrt Exp $
/* ######################################################################

 DUDF - Distribution Upgradeability Description Format

 Implementation of DUDF
 It is responsible for generating the DUDF file as defined in MANCOOSI
 project.
 DUDF contais one problem instance.
 Inspired on GlobalError class written by Jason Gunthorpe.

 ##################################################################### */
/*}}}*/
// Include Files                            /*{{{*/
#ifdef __GNUG__
#pragma implementation "apt-pkg/dudf.h"
#endif 

#include "dudf.h"

#include <libxml/parser.h>
#include <libxml/tree.h>

#include <jsoncpp/json.h>

#include <stdio.h>
//For rpm DB Dump
#include <rpm/header.h>
#include <rpm/rpmlib.h>
#include <rpm/rpmtag.h>
#include <rpm/rpmstring.h>
#include <rpm/rpmts.h>
#include <rpm/rpmds.h>
#include <rpm/rpmdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fstream>
#include <iostream>
#include <sstream>
#include <set>
#include <dirent.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <locale.h>

#include <openssl/md5.h>
#include <sys/mman.h>

#define METAINSTALLER "urpmi"
#define VERSION "8.6" 
#define RPM_VERSION_RAW "4.6.0"
/*}}}*/

using namespace std;

// Dudf Object                          /*{{{*/

GlobalDudf *_GetDudfObj() {
    static GlobalDudf *Obj = new GlobalDudf;
    return Obj;
}

string SafeGetCWD() {
    // Stash the current dir.
    char S[300];
    S[0] = 0;
    if (getcwd(S, sizeof(S) - 2) == 0)
        return "/";
    unsigned int Len = strlen(S);
    S[Len] = '/';
    S[Len + 1] = 0;
    return S;
}

/*}}}*/

// GlobalDudf::GlobalDudf - Constructor             /*{{{*/
// ---------------------------------------------------------------------
/* */
GlobalDudf::GlobalDudf() {
    time_t Date;
    struct tm * timeinfo;
    char buffer[80];

    setlocale(LC_TIME, "en_US");

    time(&Date);
    timeinfo = localtime(&Date);

    strftime(buffer, 80, "%Y%m%d%H%M%S", timeinfo);

    DudfErrorMsg = "";
    FilenameFingerprint = buffer;

    strftime(buffer, 80, "%a, %d %b %Y %H:%M:%S GMT", timeinfo);
    Timestamp = buffer;

    dudfDoc = NULL;

    PtrPkgInfo = new vector<DudfPackageInfo> ();
    //rpmdata = RPMPackageData::Singleton();

}

// GlobalDudf::~GlobalDudf - Destructor             /*{{{*/
// ---------------------------------------------------------------------
/* */
GlobalDudf::~GlobalDudf() {
    xmlFreeDoc(dudfDoc);
}

Json::Value getTagArray(Header hdr, rpmTag tagname) {
    rpmtd tagdata = rpmtdNew();
    const char * str = NULL;
    Json::Value *arr = new Json::Value(Json::arrayValue);

    if (headerGet(hdr, tagname, tagdata, HEADERGET_RAW)) {
        while ((str = rpmtdNextString(tagdata)) != NULL) {
            arr->append(Json::Value(str));
        }

    }
    rpmtdFree(tagdata);
    return *arr;

}

Json::Value getTagInteger(Header hdr, rpmTag tag) {
    rpmtd tagdata = rpmtdNew();
    Json::Value val;
    if (headerGet(hdr, tag, tagdata, HEADERGET_RAW)) {
        val = Json::Value(Json::UInt(rpmtdGetNumber(tagdata)));
    }
    rpmtdFree(tagdata);
    return val;
}

Json::Value getTagString(Header hdr, rpmTag tag) {
    rpmtd tagdata = rpmtdNew();
    Json::Value val;
    if (headerGet(hdr, tag, tagdata, HEADERGET_RAW)) {
        val = Json::Value(rpmtdGetString(tagdata));
    }
    rpmtdFree(tagdata);
    return val;
}

bool isDepFlag(uint64_t value) {
    return (value & RPMSENSE_LESS || value & RPMSENSE_GREATER || value
            & RPMSENSE_EQUAL);
}

/*
 * Generic getter for versioned dependency style info
 * It should work for requires, provides, conflicts and obsoletes
 * TODO: Maybe encode the requires as [name, flag, version] arrays as suggested by Pietro
 * see https://gforge.info.ucl.ac.be/tracker/?func=detail&aid=262&group_id=35
 *
 */
Json::Value getTagRequires(Header hdr, rpmTag tag, rpmTag flag_tag,
        rpmTag version_tag) {
    rpmtd tagdata1 = rpmtdNew();
    rpmtd tagdata2 = rpmtdNew();
    rpmtd tagdata3 = rpmtdNew();
    Json::Value *arr = new Json::Value(Json::arrayValue);
    const char * dep;
    const char * version;
    char * dep_flag = NULL;
    set<const char *, StringComparator> depSet;

    headerGet(hdr, version_tag, tagdata3, HEADERGET_RAW);
    headerGet(hdr, tag, tagdata1, HEADERGET_RAW);

    if (headerGet(hdr, flag_tag, tagdata2, HEADERGET_RAW)) {
        uint64_t depflag = 0;
        int count = 0;
        while (rpmtdNext(tagdata2) != -1) {
            depflag = rpmtdGetNumber(tagdata2);
            count++;
            rpmtdNext(tagdata1); //Requirename
            rpmtdNext(tagdata3); //Requireversion
            dep = rpmtdGetString(tagdata1);
            if (strstr(dep, "rpmlib(") == dep) //Starts with "rpmlib("
                continue;

            if (isDepFlag(depflag)) {
                dep_flag = rpmtdFormat(tagdata2, RPMTD_FORMAT_DEPFLAGS, NULL);
                version = rpmtdGetString(tagdata3);
                char *full_dep = NULL;
                rasprintf(&full_dep, "%s %s %s", dep, dep_flag, version);
                depSet.insert(full_dep);

            } else
                depSet.insert(dep);

        }

        //Using depSet assures no duplicates in a set of Deps
        set<const char *>::iterator it;
        for (it = depSet.begin(); it != depSet.end(); it++)
            arr->append(Json::Value(*it));
    }
    rpmtdFree(tagdata1);
    rpmtdFree(tagdata2);
    rpmtdFree(tagdata3);

    return *arr;

}

/**
 * Dump the RPM installation status into a JSON object to be embedded in the DUDF
 * TODO: document the json format used
 */
string GlobalDudf::dumpRPMDb() {
    stringstream rpmDbDump;
    //TODO: We should be reusing the RPM db connection that apt already has, maybe
    // through RPMDBHandler class
    rpmReadConfigFiles(NULL, NULL);
    rpmts ts = rpmtsCreate();
    Header hdr;

    rpmtsSetRootDir(ts, "/");
    //rpmtsSetVSFlags(s->ts, );
    Json::Value *array_of_pkgs = new Json::Value(Json::arrayValue);

    if (rpmtsGetRdb(ts) == NULL) {
        int rc = rpmtsOpenDB(ts, O_RDONLY);

	if (rc != 0) {
		cerr << "Error while opening RPM database. Return code: " << rc << endl;
		return "";
	}
    }

    rpmdbMatchIterator Iter = rpmtsInitIterator(ts, RPMTAG_NAME, NULL, 0);
    while ((hdr = rpmdbNextIterator(Iter)) != NULL) {
        Json::Value name = getTagString(hdr, RPMTAG_NAME);
        //Skipping RPM pseudo-packages for GPG pubkeys
        if (strcmp(name.asCString(), "gpg-pubkey") == 0)
            continue;

        Json::Value *pkg = new Json::Value(Json::arrayValue);

        pkg->append(name);
        pkg->append(getTagInteger(hdr, RPMTAG_EPOCH));
        pkg->append(getTagString(hdr, RPMTAG_VERSION));
        pkg->append(getTagString(hdr, RPMTAG_RELEASE));
        pkg->append(getTagRequires(hdr, RPMTAG_REQUIRENAME,
                RPMTAG_REQUIREFLAGS, RPMTAG_REQUIREVERSION));
        pkg->append(getTagRequires(hdr, RPMTAG_PROVIDENAME,
                RPMTAG_PROVIDEFLAGS, RPMTAG_PROVIDEVERSION));
        pkg->append(getTagRequires(hdr, RPMTAG_CONFLICTNAME,
                RPMTAG_CONFLICTFLAGS, RPMTAG_CONFLICTVERSION));
        pkg->append(getTagRequires(hdr, RPMTAG_OBSOLETENAME,
                RPMTAG_OBSOLETEFLAGS, RPMTAG_OBSOLETEVERSION));
        pkg->append(getTagInteger(hdr, RPMTAG_SIZE));
        const string &tmp = name.asString();
        pkg->append(Json::Value(isEssentialPackage(tmp)));

        array_of_pkgs->append(*pkg);
    }

    //SL: bool pretty_json = _config->FindB("APT::Dudf::PrettyJson", false);
	bool pretty_json = false;

    if (pretty_json) {
        Json::StyledWriter pretty_writer;
        rpmDbDump << pretty_writer.write(*array_of_pkgs) << endl;
    } else {
        Json::FastWriter fast;
        rpmDbDump << fast.write(*array_of_pkgs) << endl;
    }

    return rpmDbDump.str();

}

bool GlobalDudf::generateUUID(string &uuid) {

    FILE *rpmqstream;
    int UUIDSIZE = 37; // 36 chars for uuid + \0
    char tmpbuffer[UUIDSIZE];

    rpmqstream = popen("/usr/bin/uuidgen -t", "r");

    if (rpmqstream == NULL)
        //SL: return _error->Errno("generateUUID()", _("Unable to popen() uuidgen"));
        return false;

    if (fgets(tmpbuffer, UUIDSIZE, rpmqstream) == NULL)
        uuid = "";

    uuid = string(tmpbuffer);
    return true;

}

bool GlobalDudf::isEssentialPackage(const string &packageName) {
    /*return (rpmdata->PkgFlags(packageName) & pkgCache::Flag::Essential)
            == pkgCache::Flag::Essential || (rpmdata->PkgFlags(packageName)
            & pkgCache::Flag::Important) == pkgCache::Flag::Important;
            */
    return false;

}

/*
 * Retrieve and parse official repos from config files
 */
void getOfficialRepos(vector<string> &repos) {
    //SL:string repos_option = _config->Find("APT::Dudf::OfficialRepos", "");
    string repos_option = "";
    string delimiters = ";";

    // Skip delimiters at beginning.
    string::size_type lastPos = repos_option.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos = repos_option.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos) {
        // Found a token, add it to the vector.
        repos.push_back(repos_option.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = repos_option.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = repos_option.find_first_of(delimiters, lastPos);
    }
}

const char* md5OfFile(string Filename) {
    void *buffer;
    struct stat Buf;
    unsigned char result[MD5_DIGEST_LENGTH];
    char *md5 = new char[33];

    int fd_list = open(Filename.c_str(), O_RDONLY);
    if (fd_list == -1) {
	return false;
    }

    if (fstat(fd_list, &Buf) != 0) {
        cerr << "DUDF, fstat. Unable to determine the file size" << endl;
        return false;
    
    }

    buffer = mmap(NULL, Buf.st_size, PROT_READ, MAP_SHARED, fd_list, 0);
    if (!buffer) {
        cerr << "DUDF, mmap. Cannot mmap()." << endl;
        return false;
    }

    MD5((unsigned char *)buffer, Buf.st_size, result);

    for (int i = 0; i < MD5_DIGEST_LENGTH; i++) {
        sprintf(md5 + 2*i, "%02x", result[i]);
    }
    md5[32] = '\0';

    return md5;
}

//  GlobalDudf::DumpToFile- Dump DUDF into XML file     /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::DumpToFile(string Request, string ResultStatus) {

	//SL: string DudfDirName = _config->Find("Dir::Dudf");
	//if (DudfDirName == "")
	//    DudfDirName = "/var/lib/apt/dudf";
	string DudfDirName = "/var/cache/urpmi/dudf/";

	string dudfFilename = DudfDirName + "/dudf_" + FilenameFingerprint + ".xml";

	if (GlobalDudf::TryCreateDudfDocument(Request, ResultStatus))
		xmlSaveFormatFileEnc( dudfFilename.c_str() , dudfDoc, "UTF-8", 1);
	else
		cerr << "Error while generating DUDF file" << endl;

	return true;
}
/*}}}*/

//  GlobalDudf::GetFilenameFingerprint- Accessor for the FilenameFingerprint attribute /*{{{*/
// ---------------------------------------------------------------------
/* */
string GlobalDudf::GetFilenameFingerprint() {
	return FilenameFingerprint;
}
/*}}}*/

//  GlobalDudf::CreateDudf- Retrieve the XML file in a string /*{{{*/
// ---------------------------------------------------------------------
/* */
string GlobalDudf::CreateDudf(string Request, string ResultStatus) {
	xmlChar* buf;
	int length = 0;

	if (GlobalDudf::TryCreateDudfDocument(Request, ResultStatus))
		xmlDocDumpFormatMemoryEnc(dudfDoc, &buf, &length, "UTF-8", 1);
	else
		cerr << "Error while generating DUDF file" << endl;

	string xmlFile((const char *)buf);
	xmlFree(buf);

	return xmlFile;
}
/*}}}*/

//  GlobalDudf::TryCreateDudfDocument- Create document if needed /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::TryCreateDudfDocument(string Request, string ResultStatus) {
	GlobalDudf::CreateDudfDocument(Request, ResultStatus);
	return (dudfDoc != NULL);
}
/*}}}*/

//  GlobalDudf::CreateDudfDocument- Create the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::CreateDudfDocument(string Request, string ResultStatus) {
	string pkgstatus;

	string UUID;
	generateUUID(UUID);
	xmlDocPtr doc = NULL;
	xmlNodePtr root_node = NULL, node = NULL, node1 = NULL, node2 = NULL,
		   node3 = NULL, node4 = NULL;
	xmlNsPtr ns = NULL;

	doc = xmlNewDoc(BAD_CAST "1.0");
	root_node = xmlNewNode(ns , BAD_CAST "dudf");
	ns = xmlNewNs ( root_node, BAD_CAST "http://www.mancoosi.org/2008/cudf/dudf", BAD_CAST "dudf");
	/*
	 * Define DUDF version according to the RelaxNG specifications.
	 */
	xmlNewProp(root_node, BAD_CAST "dudf:version", BAD_CAST DUDF_VERSION);
	xmlDocSetRootElement(doc, root_node);

	/* 
	 * Timestamp
	 */
	/* Time is the same used for fingerprinting files */
	node1 = xmlNewChild(root_node, NULL, BAD_CAST "timestamp", BAD_CAST Timestamp.c_str());

	/*
	 * UID
	 */
	node1 = xmlNewChild(root_node, NULL, BAD_CAST "uid", BAD_CAST UUID.c_str());

	/* 
	 * The Distribution node and content
	 */

	SetDistro();

	xmlNewChild(root_node, NULL, BAD_CAST "distribution",
			BAD_CAST DudfInfoElement.Distribution.c_str());

	/* 
	 * The installer  node. 
	 */
	node = xmlNewChild(root_node, NULL, BAD_CAST "installer", NULL);
	node1 = xmlNewChild(node, NULL, BAD_CAST "name", BAD_CAST "rpm");
	node1 = xmlNewChild(node, NULL, BAD_CAST "version", BAD_CAST RPM_VERSION_RAW );

	/* 
	 * The meta-installer  node
	 */
	node = xmlNewChild(root_node, NULL, BAD_CAST "meta-installer", NULL);
	node1 = xmlNewChild(node, NULL, BAD_CAST "name", BAD_CAST METAINSTALLER);
	node1 = xmlNewChild(node, NULL, BAD_CAST "version", BAD_CAST VERSION);

	/* 
	 * The problem node
	 */
	node = xmlNewChild(root_node, NULL, BAD_CAST "problem", NULL);

	if (ResultStatus != "success") {

		/*  The package status                              */
		node1 = xmlNewChild(node, NULL, BAD_CAST "package-status", NULL);
		node2 = xmlNewChild(node1, NULL, BAD_CAST "installer", NULL);

		pkgstatus = dumpRPMDb();
		if (pkgstatus == "") {
			cerr << "Cannot get RPM database status. Aborting DUDF generation." << endl;
			return false;
		}
		const char * status = pkgstatus.c_str();

		node3 = xmlNewChild(node2, NULL, BAD_CAST "status", NULL);

		//SL: string PkgDbFilename = _config->Find("Dir::State::status",
		//        "/var/lib/rpm/Packages").c_str();
		string PkgDbFilename = "/var/lib/rpm/Packages";

		xmlNewProp(node3, BAD_CAST "filename", BAD_CAST PkgDbFilename.c_str());

		node4 = xmlNewCDataBlock(doc, BAD_CAST status, xmlStrlen(BAD_CAST status));

		xmlAddChild(node3, node4);

	}

	vector<string> *PtrListFiles = new vector<string> ();
	if ((ResultStatus != "success") && PackageUniverse(PtrListFiles)) {
		/*  The package Universe*/
		node1 = xmlNewChild(node, NULL, BAD_CAST "package-universe", NULL);

		vector<string> official_repos;
		getOfficialRepos(official_repos);

		for (vector<string>::iterator I = PtrListFiles->begin(); I
				!= PtrListFiles->end(); I++) {
			string NameFileFd = "/var/lib/urpmi/";
			node2 = xmlNewChild(node1, NULL, BAD_CAST "package-list", NULL);

			if (strstr(I->c_str(), "synthesis.hdlist") != NULL)
				xmlNewProp(node2, BAD_CAST "dudf:format", BAD_CAST "synthesis_hdlist" );
			else if (strstr(I->c_str(), "hdlist") != NULL)
				xmlNewProp(node2, BAD_CAST "dudf:format", BAD_CAST "hdlist" );

			xmlNewProp(node2, BAD_CAST "dudf:filename", BAD_CAST (NameFileFd + *I).c_str());

			bool is_official_repo = official_repos.size() > 0 ? false : true; //Ignore this check if not properly configured
			for (vector<string>::iterator Repo = official_repos.begin(); Repo
					!= official_repos.end(); Repo++) {
				//Make sure it's an official repo to find if we should include just a reference for the repo list
				if (I->find(Repo->c_str()) != string::npos) {
					is_official_repo = true;
					break;
				}
			}

			/* Not available. Need to parse source lists */
			/* xmlNewProp(node2, BAD_CAST "format", BAD_CAST ListsFilename.c_str()); */

			//SL: string NameFileFd = "/" + _config->FindDir("Dir::State::lists")
			//        + I->c_str();

			if (is_official_repo) {
				node3 = xmlNewChild(node2, NULL, BAD_CAST "hash", BAD_CAST md5OfFile(NameFileFd + *I));
			}
			else
			{
				int fd_list = open(NameFileFd.c_str(), O_RDONLY);
				if (fd_list != -1)
				{
					struct stat Buf;
					if (fstat(fd_list,&Buf) != 0) {
						return cerr << "Error in reading" << endl;
					}

					size_t file_size = Buf.st_size;

					char * list_content = (char *)malloc(file_size+1);

					if (list_content == NULL)
						cerr << "DUDF: malloc failed for list_content!";

					if (read(fd_list, list_content, file_size) == -1)
						cerr << "DUDF, read Unable to read file " << NameFileFd.c_str() << endl;

					list_content[file_size] = 0;

					node4 = xmlNewCDataBlock(doc, BAD_CAST list_content, xmlStrlen(BAD_CAST list_content));
					free(list_content);

					xmlAddChild(node2, node4);

				}

			}

		}

	}
	delete PtrListFiles;

	//string CmdCall="apt-get";
	//for (const char **I = CmdL.FileList; *I != 0; I++)
	//CmdCall= CmdCall +" " + *I;

	/*  The package universe */
	node1 = xmlNewChild(node, NULL, BAD_CAST "action", BAD_CAST Request.c_str());

	node1 = xmlNewChild(node, NULL, BAD_CAST "desiredata", NULL);
	addUrpmiSkipList(doc, node1);
	addUrpmiInstList(doc, node1);
	addUrpmiPreferList(doc, node1);
	addUrpmiPreferVendorList(doc, node1);

	/*
	 * The outcome node
	 */
	node = xmlNewChild(root_node, NULL, BAD_CAST "outcome", NULL);

	/*  Packages */
	for (vector<DudfPackageInfo>::iterator I = PtrPkgInfo->begin(); I != PtrPkgInfo->end(); I++) {

		if (I->Status == "apparently installable") {
			node1 = xmlNewChild(node, NULL, BAD_CAST "package", BAD_CAST I->Status.c_str());
			xmlNewProp(node1, BAD_CAST "name", BAD_CAST I->Name.c_str());
			xmlNewProp(node1, BAD_CAST "version-release", BAD_CAST I->VerRel.c_str());

		}
		else
		{
			node1 = xmlNewChild(node, NULL, BAD_CAST "package", NULL);
			xmlNewProp(node1, BAD_CAST "name", BAD_CAST I->Name.c_str());
			xmlNewProp(node1, BAD_CAST "version-release", BAD_CAST I->VerRel.c_str());
			node2 = xmlNewChild(node1, NULL, BAD_CAST I->Status.c_str(), NULL);
			xmlNewProp(node2, BAD_CAST "name", BAD_CAST BAD_CAST I->DepBrokenName.c_str());
			xmlNewProp(node2, BAD_CAST "constrain", BAD_CAST BAD_CAST I->DepConstrain.c_str());
			xmlNewProp(node2, BAD_CAST "version", BAD_CAST BAD_CAST I->DepConstVersion.c_str());
			node3 = xmlNewChild(node2, NULL, BAD_CAST "package" , NULL);
			xmlNewProp(node3, BAD_CAST "name", BAD_CAST I->DepBrokenName.c_str());
			if (! I->DepCandidate.empty())
				xmlNewProp(node3, BAD_CAST "candidate", BAD_CAST I->DepCandidate.c_str());
		}

	}

	/*  Result */
	//Normalize result according to DUDF schema
	if (ResultStatus != "success")
		ResultStatus = "failure";
	xmlNewProp(node, BAD_CAST "dudf:result", BAD_CAST ResultStatus.c_str());

	/*  Error msg - not printed on sucess */
	if (! DudfErrorMsg.empty())
		node1 = xmlNewChild(node, NULL, BAD_CAST "error", BAD_CAST DudfErrorMsg.c_str());

	dudfDoc = doc;
	return true;
}
/*}}}*/

//  GlobalDudf::addUrpmiFile - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiFile(xmlDocPtr doc, xmlNodePtr node, string nodeName, string filename) {
	char *buffer;
	int length;
	string buf, line;
	ifstream input;
	xmlNodePtr dataNode, cData;

	input.open(filename.c_str());
	if (!input.is_open()) {
		return false;
	}

	input.seekg(0, ios::end);
	length = input.tellg();
	input.seekg(0, ios::beg);

	buffer = new char [length];
	if (!buffer) {
		input.close();
		return false;
	}

	input.read(buffer, length);
	input.close();

	dataNode = xmlNewChild(node, NULL, BAD_CAST nodeName.c_str(), NULL); 
	cData = xmlNewCDataBlock(doc, BAD_CAST buffer, xmlStrlen(BAD_CAST buffer));
	xmlNewProp(dataNode, BAD_CAST "dudf:filename", BAD_CAST filename.c_str());
	xmlAddChild(dataNode, cData);

	delete[] buffer;
	return true;
}
/*}}}*/

//  GlobalDudf::addUrpmiSkipList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiSkipList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_skip", "/etc/urpmi/skip.list");
}
/*}}}*/

//  GlobalDudf::addUrpmiInstList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiInstList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_inst", "/etc/urpmi/inst.list");
}
/*}}}*/

//  GlobalDudf::addUrpmiPreferList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiPreferList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_prefer", "/etc/urpmi/prefer.list");
}
/*}}}*/

//  GlobalDudf::addUrpmiPreferVendorList - Add an URPMI file to the XML tree /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::addUrpmiPreferVendorList(xmlDocPtr doc, xmlNodePtr node) {
	return addUrpmiFile(doc, node, "urpmi_prefer_vendor", "/etc/urpmi/prefer.vendor.list");
}
/*}}}*/


//  GlobalDudf::SetDistro - Set distribution information        /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::SetDistro() {

    string File = "/etc/release";

    // See if the file exists
    struct stat Buf;
    if (stat(File.c_str(), &Buf) != 0)
        //SL_error->Error(_("/etc/release does not exist:"));
        cerr << "/etc/release does not exist:";

    // Open the stream for reading
    ifstream F(File.c_str(), ios::in);

    if (!F != 0)
        //_error->Error(_("Could not open /etc/release file:"));
        cerr << "Could not open /etc/release file:";

    char Buffer[1024];

    int CurLine = 0;
    while (F.eof() == false) {
        F.getline(Buffer, sizeof(Buffer));
        CurLine++;

        if (F.fail() && !F.eof())
            cerr << "Error reading /etc/release: ";

        if (strlen(Buffer) == 0)
            break;

        DudfInfoElement.Distribution = Buffer;

    }

    return true;
}

//  GlobalDudf::PackageDb - Include intensional / extensional Package DB        /*{{{*/
// ---------------------------------------------------------------------
// **DEPRECATED**
/* */
bool GlobalDudf::PackageDb(string &pkgstatus) {

    // I have to execute this process and store the result in JSON inside the  field
    //  /usr/lib/rpm/rpmq -qa --qf [\\[\"%{NAME}\",\"] --qf [%{EPOCH}:]  --qf [%{VERSION}\",\"] --qf [%{RELEASE}\",\\[] --qf [\"d\":\"%{REQUIRENAME}%{REQUIREFLAGS:depflags}%{REQUIREVERSION}\"\,]    --qf [\\],\"%{RPMVERSION}\",\\[]  --qf [\"%{PROVIDENAME}%\|PROVIDEFLAGS?{%{PROVIDEFLAGS:depflags}%{PROVIDEVERSION}}:{}\|\",]  --qf [\\],\"%{ARCH}\"]  --qf [\"%{CONFLICTNAME}%\|CONFLICTFLAGS?{%{CONFLICTFLAGS:depflags}%{CONFLICTVERSION}}:{}\|\",]   --qf [\\],\"%{INSTALLTID}\",\\]\\n]

    /* JSON Definition
     { "package-status" : [ ["name","epoch:version","release",["dep2,"dep2",..],"rpm_version",["prov1","prov2",...] ],[pkg_2],...,[pkg_n]] } */

    FILE *rpmqstream;
    char tmpbuffer[MAXBUFFER];

    pkgstatus = " { \"package-status\" : [ ";

    rpmqstream
            = popen(
                    "/bin/rpm -qa --qf [\\\\[\\\"%{NAME}\\\",\\\"]  --qf [%{EPOCH}:]  --qf [%{VERSION}\\\",\\\"] --qf [%{RELEASE}\\\",\\\\[] --qf [\\\"%{REQUIRENAME}%{REQUIREFLAGS:depflags}%{REQUIREVERSION}\\\",] --qf [\\\\],\\\"%{RPMVERSION}\\\",\\\\[]  --qf [\\\"%{PROVIDENAME}%\\|PROVIDEFLAGS?{%{PROVIDEFLAGS:depflags}%{PROVIDEVERSION}}:{}\\|\\\",]  --qf [\\\\],\\\"%{ARCH}\\\",\\\\[]    --qf [\\\"%{CONFLICTNAME}%\\|CONFLICTFLAGS?{%{CONFLICTFLAGS:depflags}%{CONFLICTVERSION}}:{}\\|\\\",]  --qf [\\\\],\\\"%{INSTALLTID}\\\"\\\\],\\\\n]  ",
                    "r");

    if (rpmqstream == NULL)
        return cerr << "Unable to get the status of the packages in the system";

    while (fgets(tmpbuffer, MAXBUFFER, rpmqstream))
        pkgstatus += tmpbuffer;

    string::iterator it = pkgstatus.end() - 2; /* That sneaky last comma... */
    pkgstatus.erase(it);

    pkgstatus += "]}";

    pclose(rpmqstream);
    /*
     * agrr:  Hack alert!: We should be using a proper JSON library
     */
    string copy = pkgstatus;
    unsigned int i;
    for (i = 0; i < copy.length() - 2; i++) {
        if (copy[i] == ',' && copy[i + 1] == ']' && copy[i + 2] == ',') /* Wooohooo */
            copy[i] = ' ';
    }
    pkgstatus = copy;

    return true;
}
/*}}}*/

//  GlobalDudf::PackageUniverse - Include intensional / extensional Package Lists       /*{{{*/
// ---------------------------------------------------------------------
/* */
bool GlobalDudf::PackageUniverse(vector<string> *PtrListFiles) {

    //string DirLists = "/" + _config->FindDir("Dir::State::lists");
    string DirLists = "/var/lib/urpmi/";

    DIR *D = opendir(DirLists.c_str());
    if (D == 0)
        cerr << "Unable to read " << DirLists.c_str();

    string StartDir = SafeGetCWD();
    if (chdir(DirLists.c_str()) != 0) {
        closedir(D);
        cerr << "Unable to change to " << DirLists.c_str();
    }

    for (struct dirent *Dir = readdir(D); Dir != 0; Dir = readdir(D)) {
        // Skip files that aren't lists
	if (strcmp(Dir->d_name, ".") == 0 || strcmp(Dir->d_name, "..") == 0)
	    continue;

	DIR *SD = opendir(Dir->d_name);
	if (SD == NULL) {
		if (errno == ENOTDIR) {
			continue;
		}

		cerr << "Unable to read " << Dir->d_name << endl;
		break;
	}
	string curDir = SafeGetCWD();
	if (chdir(Dir->d_name) != 0) {
		closedir(SD);
		cerr << "Unable to change forward to " << Dir->d_name << endl;
		break;
	}
	for (struct dirent *Subdir = readdir(SD); Subdir != 0; Subdir = readdir(SD)) {
		if (strcmp(Subdir->d_name, "lock") == 0 || strcmp(Subdir->d_name, "partial")
			== 0 || strcmp(Subdir->d_name, ".") == 0 || strcmp(Subdir->d_name,
			"..") == 0 || strstr(Subdir->d_name, "hdlist") == NULL)
		    continue;

		PtrListFiles->push_back(string(string(Dir->d_name) + "/" + string(Subdir->d_name)).c_str());
	}

        //Args[++argcount] = Dir->d_name;
	closedir(SD);
	if (chdir(curDir.c_str()) != 0) {
		cerr << "Unable to change back to " << Dir->d_name;
		break;
	}
    };

    /*
     * This code was when lists were included in a separate file

     // Fork tar
     int Process = fork();
     const char *Args[PtrListFiles->size()];

     if (Process < 0)
     return _error->Errno("fork","Couldn't fork tar");

     // The child
     if (Process == 0)
     {
     // tar util existence should be ckecked...
     string tarcmd="/bin/tar";

     FileFd FdSilentCout("/dev/null" , FileFd::WriteEmpty);
     dup2(FdSilentCout.Fd(),STDOUT_FILENO);
     SetCloseExec(STDOUT_FILENO,false);

     string ListsFilename=_config->Find("Dir::Dudf") + "/lists_" + FilenameFingerprint  + ".tar.bz2";

     Args[0] = "/bin/tar";
     Args[1] = "cvfj";
     Args[2] = ListsFilename.c_str();


     int j=3;

     for (vector<string>::iterator I = PtrListFiles->begin(); I != PtrListFiles->end(); I++,j++)
     Args[j]=I->c_str();


     Args[j] = 0;
     execvp(Args[0],(char **)Args);
     exit(100);
     }

     // Wait for tar to finish
     if (ExecWait(Process,"/bin/tar",false) == false)
     return _error->Error("tarbal of lists failed, perhaps the disk is full.");
     */

    if (chdir(StartDir.c_str()) != 0)
        return false;
    closedir(D);
    return true;

}
/*}}}*/

//  GlobalDudf::AddMsgError - Append to msg error   /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::AddMsgError(string MsgErrorPart) {

    DudfErrorMsg += MsgErrorPart;

}
/*}}}*/

//  GlobalDudf::PackagesInfo - Add info about a package /*{{{*/
// ---------------------------------------------------------------------
/* */
void GlobalDudf::PackagesInfo(string PkgName, string PkgVerRel,
        string PkgStatus, string PkdDepBrokenName, string PkgDepConstrain,
        string PkgDepConstVersion, string PkgDepCandidate) {

    DudfPackageInfo PkgElement;

    PkgElement.Name = PkgName;
    PkgElement.VerRel = PkgVerRel;
    PkgElement.Status = PkgStatus;
    PkgElement.DepBrokenName = PkdDepBrokenName;
    PkgElement.DepConstrain = PkgDepConstrain;
    PkgElement.DepConstVersion = PkgDepConstVersion;
    PkgElement.DepCandidate = PkgDepCandidate;

    PtrPkgInfo->push_back(PkgElement);

}
/*}}}*/

